from  sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import RandomForestClassifier as RF
import sklearn
import numpy as np
from sklearn.cluster import KMeans
from sklearn.metrics import adjusted_rand_score, f1_score
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import cross_val_score
import logging
import ubergauss.tools as ut


def _repeat_as_column(a,n):
    return np.tile(a,(n,1)).T

def knn_accuracy(X, y,n_neighbors):
    knn = KNeighborsClassifier(n_neighbors=n_neighbors+1)

    y_train= np.array(y)
    knn.fit(X,y)

    # Get the labels of the nearest neighbors for each training point
    _, indices = knn.kneighbors(X)
    # instances -> labels
    neighbor_labels = y_train[indices]
    # Exclude the label of the training point itself
    neighbor_labels = neighbor_labels[:, 1:]

    # Compute the training error
    agreement = (_repeat_as_column(y_train,n_neighbors) == neighbor_labels).mean()

    return agreement


def kmeans_ari(X,y,k=20):
    means = KMeans(n_clusters = k)
    predicted = means.fit_predict(X)
    return adjusted_rand_score(y,predicted)

from yoda import ml
from yoda import draw

def kmeans_bla(dist,labels):
    X = ml.embed(dist, n_dim = 2)
    draw.scatter(X,labels)
    adjusted_rand_score( KMeans(n_clusters=len(np.unique(labels))).fit_predict(X), labels)


def knn_f1(X,y,n_neighbors = 3,cv_strati_splits = 3):

    knn = KNeighborsClassifier(n_neighbors=n_neighbors)
    skf = StratifiedKFold(n_splits=cv_strati_splits, random_state=23, shuffle=True)
    y= np.array(y)
    train_test = list(skf.split(y,y))

    real = []
    pred = []
    for train, test in train_test:
        knn.fit(X[train],y[train])
        pred += list(knn.predict(X[test]))
        real += list(y[test])

    logging.captureWarnings(True)
    logging.captureWarnings(False)
    calc_jointly =  f1_score(real,pred,average=f'weighted')
    # mean of f1 scores as in rnascclust paper
    return np.mean( cross_val_score(knn, X, y, cv=skf, scoring=f'f1_weighted', n_jobs=1))


def overlap_coef(csr_a, csr_b, nonorm=False):
    norm = 1 if nonorm else min(len(csr_a.indices), len(csr_b.indices))
    score =  len(np.intersect1d(csr_a.indices, csr_b.indices)) / norm
    # score = np.intersect1d(a, b)/ min(len(a), len(b))
    return 1-score

def overlap(csrs, nonorm =False):
    s = csrs.shape[0]
    m = np.zeros((s,s))
    for i in range(s):
        for j in range(i,s):
            m[i,j] = m[j,i] =overlap_coef(csrs[i], csrs[j],nonorm=nonorm)
    return m

def overlap2(csrs):
    s = csrs.shape[0]
    def dorow(i):
        m = np.zeros(s)
        for j in range(i,s):
            m[j] = overlap_coef(csrs[i], csrs[j])
        return m

    return np.vstack(ut.xmap(dorow,range(s)))

def featuremask(X,y,n_ft = 100):
    # clf = RF(n_estimators = 100,max_features=None).fit(X,y)
    clf = RF(n_estimators = 100).fit(X,y)
    featscores =  clf.feature_importances_
    # print(f"{ featscores=}")
    return ut.binarize(featscores, n_ft)

def permutation_score(X,y):
    estimator = sklearn.cluster.KMeans(n_clusters = len(np.unique(y)))
    score = sklearn.model_selection.permutation_test_score(estimator, X, y, cv = 2)
    return score[2] # this is the pvalue



from lmz import Map,Zip,Filter,Grouper,Range,Transpose,Flatten
import ubergauss.tools as ut

import networkx.linalg as nxlin
import numpy as np
def makematrix(graph, maxlen = None):

    startpadding = 5
    endpadding = 5

    # add padding nodes
    missing_nodes = maxlen - len(graph)
    missing_nodes += endpadding # extra padding
    maxlen += startpadding+endpadding

    for i in range(missing_nodes):
        graph.add_node(max(graph.nodes)+1 , label = 'N')

    for i in range(startpadding):
        graph.add_node( -i-1 , label = 'N')

    assert len(graph)== maxlen, f"{len(graph)=}, {maxlen=}"
    sorted_nodes = sorted(graph.nodes)
    adjacency_matrix = nxlin.adjacency_matrix(graph,nodelist=sorted_nodes).todense()
    # now we need to add another axis and add node infos
    res = np.zeros((11, maxlen, maxlen))
    indices = np.repeat ( Range(maxlen), maxlen).reshape((maxlen,maxlen))
    def makeonehot(char):
        return np.repeat ( [ 1 if graph.nodes[n]['label'] == char else 0 for n in sorted_nodes ] , maxlen).reshape((maxlen,maxlen))

    ONEHOT = np.array( Map(makeonehot, 'AUGC'))
    res[:4] = ONEHOT
    res[4] = indices
    res[5] = adjacency_matrix
    res[6] = indices.T
    res[7:] = np.transpose(ONEHOT, (0,2,1))

    return res


def toarray(alis):
    graphs = [a.graph for a in alis]
    maxlen = max(map(len, graphs))
    return ut.xmap(makematrix, graphs, maxlen = maxlen)


def test_toarray():
    from yoda import alignments
    a,l = alignments.load_rfam()
    atensor = toarray(a[:2])[0]


    import structout as so
    for layer in atensor:
        so.heatmap(layer)






import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

### MNIST code originally from https://github.com/pytorch/examples/blob/master/mnist/main.py ###
from torchvision import datasets, transforms

from pytorch_metric_learning import distances, losses, miners, reducers, testers
from pytorch_metric_learning.utils.accuracy_calculator import AccuracyCalculator


### MNIST code originally from https://github.com/pytorch/examples/blob/master/mnist/main.py ###
class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = nn.Dropout2d(0.25)
        self.dropout2 = nn.Dropout2d(0.5)
        self.fc1 = nn.Linear(9216, 128)

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        return x


### MNIST code originally from https://github.com/pytorch/examples/blob/master/mnist/main.py ###
def train(model, loss_func, mining_func, device, train_loader, optimizer, epoch):
    model.train()
    for batch_idx, (data, labels) in enumerate(train_loader):
        data, labels = data.to(device), labels.to(device)
        optimizer.zero_grad()
        embeddings = model(data)
        indices_tuple = mining_func(embeddings, labels)
        loss = loss_func(embeddings, labels, indices_tuple)
        loss.backward()
        optimizer.step()
        if batch_idx % 20 == 0:
            print(
                "Epoch {} Iteration {}: Loss = {}, Number of mined triplets = {}".format(
                    epoch, batch_idx, loss, mining_func.num_triplets
                )
            )


### convenient function from pytorch-metric-learning ###
def get_all_embeddings(dataset, model):
    tester = testers.BaseTester()
    return tester.get_all_embeddings(dataset, model)


### compute accuracy using AccuracyCalculator from pytorch-metric-learning ###
def test(train_set, test_set, model, accuracy_calculator):
    train_embeddings, train_labels = get_all_embeddings(train_set, model)
    test_embeddings, test_labels = get_all_embeddings(test_set, model)
    train_labels = train_labels.squeeze(1)
    test_labels = test_labels.squeeze(1)
    print("Computing accuracy")
    accuracies = accuracy_calculator.get_accuracy(
        test_embeddings, test_labels, train_embeddings, train_labels, False
    )
    print("Test set accuracy (Precision@1) = {}".format(accuracies["precision_at_1"]))


device = torch.device("cuda")

transform = transforms.Compose(
    [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
)

batch_size = 256






dataset1 = datasets.MNIST(".", train=True, download=True, transform=transform)
dataset2 = datasets.MNIST(".", train=False, transform=transform)
train_loader = torch.utils.data.DataLoader(
    dataset1, batch_size=batch_size, shuffle=True
)
test_loader = torch.utils.data.DataLoader(dataset2, batch_size=batch_size)



model = Net().to(device)
optimizer = optim.Adam(model.parameters(), lr=0.01)
num_epochs = 1


### pytorch-metric-learning stuff ###
distance = distances.CosineSimilarity()
reducer = reducers.ThresholdReducer(low=0)
loss_func = losses.TripletMarginLoss(margin=0.2, distance=distance, reducer=reducer)
mining_func = miners.TripletMarginMiner(
    margin=0.2, distance=distance, type_of_triplets="semihard"
)
accuracy_calculator = AccuracyCalculator(include=("precision_at_1",), k=1)
### pytorch-metric-learning stuff ###


for epoch in range(1, num_epochs + 1):
    train(model, loss_func, mining_func, device, train_loader, optimizer, epoch)
    test(dataset1, dataset2, model, accuracy_calculator)



import networkx as nx
import numpy as np
from lmz import Map
from scipy import sparse
from ubergauss import tools as ut
from eden import graph as eg
import grakel
import time

def vectorize_alignments(alignments,min_rd = 1, ignorevectorlabels = False, mp = False):
    vectorizer = lambda x: eg.vectorize([x.graph],
                                        discrete = ignorevectorlabels,
                                        min_r = min_rd,
                                        min_d = min_rd) # normalization=False, inner_normalization=False)
    mapper = ut.xmap if mp else Map
    vectors = mapper(vectorizer, alignments)
    vectors = sparse.vstack(vectors)
    return vectors


def convert_to_grakel_graph(alignment) -> grakel.Graph:

    graph = alignment.graph
    dol = nx.convert.to_dict_of_lists(graph)
    node_attr = dict()
    for nid, attr in graph.nodes(data=True):
        # node_attr[nid] = np.asarray([attr['label']]+attr['vec'].tolist())
        node_attr[nid] = ord(attr[f'label'])# np.asarray([ord(attr['label'])]+attr['vec'])

    edge_attr = {}
    for e in graph.edges:
        a,b = e
        edge_attr[e] = ord(graph.edges[a,b]['label'])

    result = grakel.Graph(dol, node_labels=node_attr, edge_labels=edge_attr, graph_format='dictionary')
    for node in [n for n in dol.keys() if n not in result.edge_dictionary.keys()]:
        result.edge_dictionary[node] = {}
    return result



def grakel_vectorize(alignments, kernel= 'WeisfeilerLehman'):
    g =  Map(convert_to_grakel_graph, alignments )
    d = {'normalize' : False} # , 'sparse' : True}
    # jstarttime = time.time()
    if kernel == f'OddSth':
        d[f'h'] = 4
    res =  eval(f'grakel.{kernel}')(**d).fit_transform(g)
    # return time.time()-starttime
    return res
# GraphHopper RandomWalkLabeled
grakel_kernels = 'ShortestPath RandomWalk PyramidMatch NeighborhoodHash GraphletSampling SubgraphMatching WeisfeilerLehman HadamardCode NeighborhoodSubgraphPairwiseDistance LovaszTheta SvmTheta Propagation PropagationAttr OddSth MultiscaleLaplacian HadamardCode VertexHistogram EdgeHistogram CoreFramework WeisfeilerLehmanOptimalAssignment'.split(f' ')






import umap
from sklearn.metrics import pairwise


def embed(distances, n_dim = 2):
    return umap.UMAP(n_components=n_dim, metric='precomputed',n_neighbors=2).fit_transform(distances)


def get_distance_method(distance_measure):
    if distance_measure == 'euclidean':
        di = pairwise.euclidean_distances
    elif distance_measure == 'overlap':
        di = simpleMl.overlap
    elif distance_measure == 'overlap_nonorm':
        di = lambda x: 1 - simpleMl.overlap(x,nonorm=True)
    elif distance_measure == 'linkern':
        di = lambda x: 10 - pairwise.linear_kernel(x)
    else:
        assert False, 'distance_measure invalid'
    return di



from lmz import Map,Zip,Filter,Grouper,Range,Transpose,Flatten
import numpy as np
import glob
import networkx as nx
import eden.graph as eg # eden-kernel in pip
from collections import defaultdict, Counter
from sklearn.preprocessing import normalize
import math
import ubergauss.tools as ut
from yoda.graphs import ali2graph
import yoda.alignments.alignment as ali
import re


'''
i think we should just load alignments here...
'''

def _split_on_empty_lines(s):
    blank_line_regex = r'\r?\n\s*\n'
    return re.split(blank_line_regex, s.strip())

def read_stk_file(fname):
    text = open(fname, 'r').read()
    alignments = _split_on_empty_lines(text)
    r = Map(ali.stk_to_alignment, alignments, fname = fname)
    return [a for a in r if a]



def readseedfile(fname):
    text = open(fname, 'r').read()
    alignments = _split_on_slashstockholm(text)
    r = Map(ali.stk_to_alignment, alignments, fname = False)
    return [a for a in r if a]

def _split_on_slashstockholm(s):
    blank_line_regex = r'//\n# STOCKHOLM 1.0'
    return re.split(blank_line_regex, s.strip())



def split_on_newseq(s):
    blank_line_regex = r'>.*'
    return re.split(blank_line_regex, s.strip())


def read_fasta(fname):
    text =  open(fname, 'r').read()
    sequences = split_on_newseq(text)
    sequences = [s.strip() for s in sequences if s]
    sequences = np.array([list(a.upper()) for a in sequences])
    return ali.Alignment(sequences, {}, {}, fname)

#############################
# nanaman
#############################


# TODO, theese are not actually bad they just contain columns without nucleotides
bad = '''/home/stefan/WEINBERG//neg/550-1278861-0-0.sto
/home/stefan/WEINBERG//neg/550-1454352-0-0.sto
/home/stefan/WEINBERG//neg/550-1382269-0-0.sto
/home/stefan/WEINBERG//neg/769-1136-0-0.sto
/home/stefan/WEINBERG//neg/769-1136-0-0.sto
/home/stefan/WEINBERG//neg/550-1261710-0-0.sto
/home/stefan/WEINBERG//neg/550-1261710-0-0.sto
/home/stefan/WEINBERG//neg/550-211660-0-0.sto
/home/stefan/WEINBERG//neg/592-880-0-0.sto
/home/stefan/WEINBERG//neg/592-880-0-0.sto
/home/stefan/WEINBERG//neg/592-880-0-0.sto
/home/stefan/WEINBERG//neg/592-880-0-0.sto
/home/stefan/WEINBERG//neg/592-880-0-0.sto
/home/stefan/WEINBERG//neg/550-2011241-0-0.sto
/home/stefan/WEINBERG//neg/550-2011241-0-0.sto
/home/stefan/WEINBERG//neg/550-2025985-0-0.sto
/home/stefan/WEINBERG//neg/550-2025985-0-0.sto
/home/stefan/WEINBERG//neg/550-2025985-0-0.sto
/home/stefan/WEINBERG//neg/550-2025985-0-0.sto
/home/stefan/WEINBERG//neg/550-1260425-0-0.sto
/home/stefan/WEINBERG//neg/550-1927583-0-0.sto
/home/stefan/WEINBERG//neg/550-1927583-0-0.sto
/home/stefan/WEINBERG//neg/550-2308630-0-0.sto
/home/stefan/WEINBERG//neg/550-2308630-0-0.sto
/home/stefan/WEINBERG//neg/550-2308630-0-0.sto
/home/stefan/WEINBERG//neg/550-1708760-0-0.sto'''.split('\n')


from collections import Counter



def _getfiles_70k(path = '',removesmall=True, limit = 0):
    files = []
    for asd in 'neg pos pos2'.split():
        currentfiles  = glob.glob(f'{path}/{asd}/*')
        if not currentfiles:
            print(f'path is wrong: {path}')
        if removesmall:
            currentfiles = filtersmall(currentfiles)
        if limit:
            currentfiles = currentfiles[:limit]
        currentfiles = [c for c in currentfiles if c not in bad]
        files.append(currentfiles)
    return files

def issmall(f):
    return read_stk_file(f)[0].alignment.shape[0] < 3

def filtersmall(files):
    toosmall = ut.xmap(issmall, files)
    return [f for f, bad in zip(files, toosmall) if not bad]


def getXYFiles_70k(path = '', limit = 0, encode = 'mainchainentropy'):
    allfiles = _getfiles_70k(path = path,removesmall = True, limit = limit)
    flatfiles = Flatten(allfiles)
    alis = ut.xmap(read_stk_file, flatfiles)
    alis = Flatten(alis)
    mkvec = lambda x: eg.vectorize([ali2graph.mainchainentropy(x)])
    vectors = ut.xmap(mkvec, alis, processes = 88)
    values = [ i  for i,e in enumerate(allfiles) for z in Range(e) ]
    return vectors, values, alis

#########################################
# RFAMOME
####################################
def _mktuple(x):
    x = x.split()
    return (int(x[1])-1,int( x[2])-1, float(x[4])) #  ['~', '46', '124', '4.14138', '0.0468148', '6', '0.03']


def loadrfamome(path, verbose = True):
    currentfiles  = glob.glob(f'{path}/*.fasta')
    alis = ut.xmap(read_fasta, currentfiles)
    print(f"{Counter([ali.label for ali in alis])}")
    return alis


def addcov(alis):
    for a in alis:
        covname = a.fname.replace(f'.fasta',f'_1.cov')
        try:
            text = open(covname, 'r').read()
        except:
            text = f''
            print(f" no cov file ; { covname=}")
        allcov = re.findall(r'~.*', text)
        a.rscape = [_mktuple(f) for f in allcov]
    return alis

def addstructure(alis):
    for a in alis:
        strname = a.fname+f'.lina'
        text = open(strname, 'r').readlines()[1].strip()
        a.struct = text
    return alis



def process_cov(alis, debug = False):
    for ali in alis:
        try:
            s= ali.struct
            cov = ali.rscape
        except:
            print(f'structure and cov are missing... abort')

        stack = []
        pairs = []
        for i,e in enumerate(s):
            if e == f'(':
                stack.append(i)
            if e == f')':
                pairs.append((stack.pop(),i))
        annotation = [0]*len(s)
        for start,end,value in cov:
            if (start,end) in pairs:
                annotation[start] = value
                annotation[end] = value
        ali.covariance = annotation
        ali.pairs = pairs
        if debug:
            print(f"{ annotation}")
    return alis



################
# ijust do this on the rfam dataset....
####################
def addcov_rfam(alis, path ):
    for ali in alis:
        cname = ut.fixpath( f'{path}/{ali.label}_{ali.gf["ID"][3:]}.sorted.cov' )
        try:
            text = open(cname, 'r').read()
            allcov = re.findall(r'~.*', text)
            ali.rscape = [_mktuple(f) for f in allcov]
        except:
            print(f"no file: { cname=}, no problem there is just nothing covariing")
            ali.rscape = []
    return alis



from lmz import Map,Zip,Filter,Grouper,Range,Transpose,Flatten
from lmz import Map
from sklearn.metrics import silhouette_score
import numpy as np
from yoda.ml import simpleMl
from ubergauss import tools as ut
from yoda.graphs import ali2graph
from sklearn.metrics import pairwise
import matplotlib
from yoda import draw
# from yoda.graphs import vectorize_alignments, grakel_vectorize
import yoda.graphs as graphs
from yoda.ml import embed, get_distance_method
matplotlib.use('module://matplotlib-sixel')
import ubergauss.optimization as uo
from yoda import alignments


########################
# so, first we cluster all the right alignments
##########################
# data = alignments.load_rfam()

def eval_ft(X_new,_,y):
    # X=X.todense()
    # X = pairwise.euclidean_distances(X_new)
    # X = embed(X, n_dim= 6)
    #return silhouette_score(X_new, y, metric= f'precomputed')
    return silhouette_score(X_new, y)

def add_vector_attributes(ali):
    return  ut.xmap( lambda ali:ali2graph.rfam_graph_decoration(ali,
                       RY_thresh = .7,nuc_thresh = .65, conservation = [],
                        covariance = 0.05, sloppy = False, fake_nodes = False, progress = True,
                        nucleotideRNAFM = False),
                ali)



import metric_learn
from collections import Counter
def supervised(alis,labels, n_ft = 100):
    a = add_vector_attributes(alis)

    # i think this samples sequences
    # a,labels = Transpose( Flatten ( ut.xmap(lambda x: ali2graph.manifest_subgraphs(x, maxgraphs = 10),zip(a,labels))))
    # a, labels = Transpose(Flatten ([ (newali,label)  for ali,label in zip(a,labels) for newali in ali2graph.manifest_subgraphs(ali,100) ] ))
    # there is manifest_sequences now !
    a, labels = ali2graph.manifest_sequences(a,labels, instances = 5  )

    labels = np.array(labels)
    X = graphs.vectorize_alignments(a, min_rd=1, mp= True)

    # remove empty columns
    X=X.toarray()
    import yoda.ml.simpleMl as ml
    ft = ml.featuremask(X, labels, n_ft = 1000)
    X = X[:,ft==1]
    print(f"{X.shape=}")

    # empty_columns = np.all(X == 0, axis=0)
    # X = X[:, ~empty_columns]


    test_scores = []
    for train, test in uo.groupedCV(n_splits = 3).split(X,labels,labels):
        tr = X[train], labels[train]
        te = X[test], labels[test]
        print(f"{ Counter(tr[1])=}")
        model = metric_learn.LFDA(n_components = 6) # LFDA and MLKR should be tried
        train_transformed = model.fit_transform(*tr)
        test_transformed = model.transform(te[0])
        test_scores.append( (eval_ft(test_transformed, *te)) )
        print(f"{ eval_ft(train_transformed, *tr) = }")
    print(f"{ np.mean(test_scores)= }")


if __name__ == f"__main__":
    data = alignments.load_rfam()
    supervised(*data)


from scipy import sparse

def supervised_graphaverage(alis,labels, n_ft = 100):
    a = add_vector_attributes(alis)

    def stacksamples(ali):
        aliss, _  = Transpose(ali2graph.manifest_subgraphs((ali,False),maxgraphs = 10) )
        X = graphs.vectorize_alignments(aliss, min_rd=1, mp= False)
        return np.mean(X, axis = 0)

    X =  ut.xmap(stacksamples , a)
    X = np.vstack(X)

    print(f"{X.shape=}")
    test_scores = []
    for train, test in uo.groupedCV(n_splits = 3).split(X,labels,labels):
        tr = X[train], labels[train]
        te = X[test], labels[test]
        ft = simpleMl.featuremask(*tr, n_ft)
        test_scores.append( (eval_ft(ft, *te)) )
        print(f"{ eval_ft(ft, *tr) = }")
    print(f"{ np.mean(test_scores)= }")




import numpy as np

match = lambda x,y: 1 if x==y else -1





def smith_waterman_score(a,b, gap = -1, edgegap= -.03):

    # make some variables
    la = len(a)
    lb = len(b)
    matrix = np.zeros((la+1,lb+1))

    # # we also need to init oO
    # matrix[0] = np.arange(lb+1) * edgegap
    # matrix[:,0] = np.arange(la+1) * edgegap

    # lets fill the matrix
    for i in range(la):
        for j in range(lb):
            options = (matrix[i,j]+match(a[i], b[j]),0, # diagonal or zero ;)
                       matrix[i,j+1]+(gap if j < lb-1 else edgegap),             # down
                       matrix[i+1,j]+(gap if i < la-1 else edgegap))              # sideways
            matrix[i+1,j+1] = np.max(options)

    return np.max(matrix)

def needle(a,b, gap = -1, edgegap= -.03, get_score_only = False):

    # make some variables
    la = len(a)
    lb = len(b)
    matrix = np.zeros((la+1,lb+1))
    choices = np.zeros((la+1,lb+1))

    # we also need to init oO
    matrix[0] = np.arange(lb+1) * edgegap
    matrix[:,0] = np.arange(la+1) * edgegap
    choices [0] = 2
    choices [:,0] = 1

    # lets fill the matrix
    for i in range(la):
        for j in range(lb):
            options = (matrix[i,j]+match(a[i], b[j]), # diagonal
                       matrix[i,j+1]+(gap if j < lb-1 else edgegap),             # down
                       matrix[i+1,j]+(gap if i < la-1 else edgegap))              # sideways
            matrix[i+1,j+1] = max(options)
            choices[i+1,j+1] = np.argmax(options)

    if get_score_only:
        return matrix[la,lb]

    a = list(a)
    b = list(b)
    ca,cb = la,lb
    reta,retb = '',''
    while True:
        if ca == 0 and cb ==0:
            break

        elif choices[ca,cb] == 0:
            reta+=a.pop()
            retb+=b.pop()
            ca -=1
            cb -=1

        elif choices[ca,cb] == 1:
            reta+=a.pop()
            retb+='-'
            ca -=1

        elif choices[ca,cb] == 2:
            retb+=b.pop()
            reta+='-'
            cb -=1

    return reta[::-1], retb[::-1]









# https://github.com/pereperi/Gotoh-algorithm

def gotoh(seqI, seqJ):

    matrix = lambda x,y: 1 if x==y else -1
    lenI = len(seqI)
    lenJ = len(seqJ)

    gop = -7
    gep = -1

    m = [[0 for x in range(lenJ + 1)] for y in range(lenI + 1)]
    ix = [[0 for x in range(lenJ + 1)] for y in range(lenI + 1)]
    iy = [[0 for x in range(lenJ + 1)] for y in range(lenI + 1)]
    tbm = [[0 for x in range(lenJ + 1)] for y in range(lenI + 1)]
    tbx = [[0 for x in range(lenJ + 1)] for y in range(lenI + 1)]
    tby = [[0 for x in range(lenJ + 1)] for y in range(lenI + 1)]

    for i in range(1, lenI + 1):
        iy[i][0] = gop + (i-1) * gep
        ix[i][0] = float('-inf')
        m[i][0] = float('-inf')
        tbm[i][0] = 1
        tbx[i][0] = 1
        tby[i][0] = 1


    for j in range(1, lenJ + 1):
        ix[0][j] = gop + (j-1) * gep
        iy[0][j] = float('-inf')
        m[0][j] = float('-inf')
        tbm[0][j] = -1
        tbx[0][j] = -1
        tby[0][j] = -1

    for i in range(1, lenI + 1):
        for j in range(1, lenJ + 1):
            s = matrix(seqI[i-1], seqJ[j-1])
            #M
            sub = m[i-1][j-1] + s
            x = ix[i-1][j-1] + s
            y = iy[i-1][j-1] + s
            if sub >= x and sub >= y:
                m[i][j] = sub
                tbm[i][j] = 0
            elif x > y:
                m[i][j] = x
                tbm[i][j] = -1
            else:
                m[i][j] = y
                tbm[i][j] = 1
            #Ix
            sub = m[i][j-1] + gop
            x = ix[i][j-1] + gep
            if sub >= x:
                ix[i][j] = sub
                tbx[i][j] = 0
            else:
                ix[i][j] = x
                tbx[i][j] = -1
            #Iy
            sub = m[i-1][j] + gop
            y = iy[i-1][j] + gep
            if sub >= y:
                iy[i][j] = sub
                tby[i][j] = 0
            else:
                iy[i][j] = y
                tby[i][j] = 1


    i = lenI
    j = lenJ
    alnI = []
    alnJ = []

    if m[i][j] >= ix[i][j] and m[i][j] >= iy[i][j]:
        state = 0
    elif ix[i][j] > iy[i][j]:
        state = -1
    else:
        state = 1

    while i != 0 or j != 0:
        if state == 0:
            state = tbm[i][j]
            i += -1
            j += -1
            alnI.append(seqI[i])
            alnJ.append(seqJ[j])
        elif state == -1:
            state = tbx[i][j]
            j += -1
            alnI.append("-")
            alnJ.append(seqJ[j])
        else:
            state = tby[i][j]
            i += -1
            alnI.append(seqI[i])
            alnJ.append("-")

    seqI_aln = "".join(reversed(alnI))
    seqJ_aln = "".join(reversed(alnJ))
    return seqI_aln, seqJ_aln




import numpy as np
import re


class Alignment:
    def __init__(self, ali,gc,gf, fname = ''):
        self.alignment = ali
        self.gc = gc
        self.gf = gf
        if not fname:
            fname= gf.get(f'AC', f'no idea :) ')
        self.fname = fname
        self.label  = grepfamily(fname)

    def __repr__(self):
        return f'\n'.join([f''.join(row) for row in self.alignment])+'\n'


def grepfamily(name):
    return re.findall(r'RF\d\d\d\d\d',name)[0]



def stk_to_alignment(text, fname):
    alignment = []
    gc = {}
    gf = {}
    for line in text.split(f'\n'):
        if line.startswith( "# STOCK") or not len(line) or line.startswith("/"):
            continue
        elif line.startswith("#=GC"):
            _, name, value = line.split()
            gc[name] = value
        elif line.startswith("#=GF"):
            split = line.split()
            gf[split[1]] = ' '.join(split[1:])
        elif not line.startswith('#='):
            alignment.append(line.split()[1])
    if not alignment:
        return None
    alignment = np.array([list(a.upper()) for a in alignment])
    return Alignment(alignment, gc, gf, fname)

'''
def fasta(ali):
    out = f''
    text= []
    filename = str(uuid.uuid4())
    for i, line in enumerate(ali.alignment):
        text.append(f'>{i}')
        text.append(f''.join(line)) # this could work...
    text = f'\n'.join(text)
    return filename
'''




from lmz import Map,Zip,Filter,Grouper,Range,Transpose,Flatten
from lmz import Map
from sklearn.metrics import silhouette_score
import numpy as np
from yoda.ml import simpleMl
from ubergauss import tools as ut
from yoda.graphs import ali2graph
from sklearn.metrics import pairwise
import matplotlib
from yoda import draw
# from yoda.graphs import vectorize_alignments, grakel_vectorize
import yoda.graphs as graphs
from yoda.ml import embed, get_distance_method
matplotlib.use('module://matplotlib-sixel')
import ubergauss.optimization as uo
from yoda import alignments


########################
# so, first we cluster all the right alignments
##########################
# data = alignments.load_rfam()



def find_working_grakel_kernels(alignments,i):
    """
    this is just to figure out which kernels work in the first place
    """
    grakel_kernels = graphs.grakel_kernels
    alignments = add_vector_attributes(alignments[:2])
    graphs.grakel_vectorize(alignments,grakel_kernels[i])
    oklist = []
    for i,v in enumerate(grakel_kernels):
        try:
            dist = graphs.grakel_vectorize(alignments,v)
            # print(f"{ dist=}")
            print(f"{ v=}{i=} ok")
            oklist.append(v)
        except:
            print(f"{ v=}{i=} failed")
    print(oklist)

import structout as so
def run( alignments,labels,
        RY_thresh = .5,nuc_thresh = .5, conservation = [],
        covariance = False, sloppy = False, fake_nodes = False, progress = False,
        kernel = f'', # 'WeisfeilerLehman',
        nucleotideRNAFM = False,
        distance_measure = 'euclidean',min_rd = 2,n_dim = 6, mp = False):

    mapper = ut.xmap if mp else Map

    alignments = mapper( lambda ali:ali2graph.rfam_graph_decoration(ali,
                          RY_thresh = RY_thresh ,
                          nuc_thresh=nuc_thresh,
                          conservation = conservation,
                          nucleotideRNAFM = nucleotideRNAFM,
                          covariance = covariance,
                          sloppy = sloppy,
                          progress= progress,
                          fake_nodes = fake_nodes),
                          alignments)


    # draw.asciigraph(alignments[4])
    # draw.shownodes(alignments[4])
    # return


    if kernel:
        sim_matrix  = graphs.grakel_vectorize(alignments, kernel)
        if np.max(sim_matrix)==0:
            return -1
        distance_matrix = 1-sim_matrix/np.max(sim_matrix)
        np.fill_diagonal(distance_matrix, 0)
    else:
        vectors = graphs.vectorize_alignments(alignments, mp=mp, min_rd= min_rd)
        distance_matrix = get_distance_method(distance_measure)(vectors)

    return silhouette_score(distance_matrix, labels, metric='precomputed')

    # dist = di(vectors)
    # X = embed(dist, n_dim= n_dim)
    # # plotneighbors(alignments,dist, alignment_id = 5, num  = 4)
    # # scatter(X, labels)
    # # print(f"{ simpleMl.permutation_score(X, labels) = }")
    # return silhouette_score(X, labels)
    # # return silhouette_score(X, labels),
############
# get goodd features
#####################


def eval_ft(ft,X,y):
    # X=X.todense()
    X = X[:,ft==1]
    X = pairwise.euclidean_distances(X)
    # X = embed(X, n_dim= 6)
    return silhouette_score(X, y, metric= f'precomputed')

def add_vector_attributes(ali):
    return  ut.xmap( lambda ali:ali2graph.rfam_graph_decoration(ali,
                       RY_thresh = .7,nuc_thresh = .65, conservation = [],
                        covariance = 0.05, sloppy = False, fake_nodes = False, progress = False,
                        nucleotideRNAFM = False),
                ali)

def supervised(alis,labels, n_ft = 100):
    a = add_vector_attributes(alis)

    a,labels = Transpose( Flatten ( ut.xmap(lambda x: ali2graph.manifest_subgraphs(x,maxgraphs = 10),zip(a,labels))))
    # a, labels = Transpose(Flatten ([ (newali,label)  for ali,label in zip(a,labels) for newali in ali2graph.manifest_subgraphs(ali,100) ] ))
    labels = np.array(labels)

    X = graphs.vectorize_alignments(a, min_rd=1, mp= True)
    for n_ft in [1000,3000]:
        test_scores = []
        for train, test in uo.groupedCV(n_splits = 3).split(X,labels,labels):
            tr = X[train], labels[train]
            te = X[test], labels[test]
            ft = simpleMl.featuremask(*tr, n_ft)
            test_scores.append( (eval_ft(ft, *te)) )
            print(f"{ eval_ft(ft, *tr) = }")

        print(f"{ np.mean(test_scores)= }")

from scipy import sparse

def supervised_graphaverage(alis,labels, n_ft = 100):
    a = add_vector_attributes(alis)

    def stacksamples(ali):
        aliss, _  = Transpose(ali2graph.manifest_subgraphs((ali,False),maxgraphs = 10) )
        X = graphs.vectorize_alignments(aliss, min_rd=1, mp= False)
        return np.mean(X, axis = 0)

    X =  ut.xmap(stacksamples , a)
    X = np.vstack(X)

    print(f"{X.shape=}")
    test_scores = []
    for train, test in uo.groupedCV(n_splits = 3).split(X,labels,labels):
        tr = X[train], labels[train]
        te = X[test], labels[test]
        ft = simpleMl.featuremask(*tr, n_ft)
        test_scores.append( (eval_ft(ft, *te)) )
        print(f"{ eval_ft(ft, *tr) = }")
    print(f"{ np.mean(test_scores)= }")




############
#   grip optimization
##########
from ubergauss import optimization as opti
grid = {
        # 'RY_thresh' : np.linspace(0.4, 0.75, 10),
        # 'nuc_thresh': np.linspace(.36,.7,10),
        # 'distance_measure': 'euclidean '.split(),
        'fake_nodes': [False, True],
        'sloppy': [False,True],
        # 'min_rd': [1,2],
        'progress' : Range(2,40,10),
        f'covariance': [False, .05],
        # 'nucleotideRNAFM':[True, False],
        # 'vectorizer' : ['ShortestPath', 'PyramidMatch', 'NeighborhoodHash', 'GraphletSampling', 'WeisfeilerLehman', 'SvmTheta', 'Propagation', 'OddSth', 'VertexHistogram', 'EdgeHistogram', 'CoreFramework', 'WeisfeilerLehmanOptimalAssignment']
        }

def optimize():
    data = alignments.load_rfam(full = False)
    # data, label  = data
    df =  opti.gridsearch(run, grid, data)
    print(df.corr(method='spearman'))
    opti.dfprint(df)
    ut.dumpfile(df,f'lastopti.delme')

def drawoptidf():
    import seaborn as sns
    import matplotlib.pyplot as plt
    # sns.scatterplot(data = df, x = f'RYlimit', y = f'nuclimit', c = list(-df['score']))
    df = ut.loadfile(f'lastopti.delme')

    mydf = df.pivot(f'RY_thresh', f'nuc_thresh', f'score')
    sns.heatmap(mydf)
    plt.tight_layout()
    plt.show()
    plt.close()




from lmz import Map,Zip,Filter,Grouper,Range,Transpose,Flatten
import numpy as np
import networkx as nx
import eden.graph as eg # eden-kernel in pip
from collections import defaultdict, Counter

open_brace_string={")":"(",
                "]":"[",
                ">":"<","}":"{"}



def nested_frag_encoder(ali):
    '''
    - adds entropy info to into vectorlabels
    - builds backbone along the main structure,
            if there are side-structures (indicated by dots in ss_cons)
            they get connected via darkedges
    -> hence the name nested fragment encoder
    '''
    graph = nx.Graph()
    lifo = defaultdict(list)

    stru = ali.gc['SS_cons']
    for i, ssc in enumerate(stru):
        # FIND NODE LABEL
        # most common:
        ct = Counter( ali.alignment[:,i].tolist())
        for k,v in ct.most_common():
            if k in "ACGU":
                nodelabel = k
                break
        # ADD NODE,
        # with a vector annotation,,, am i using that?
        myv = [ord(ali.gc[k][i]) for k in ali.gc.keys()]
        graph.add_node(i, label=k, vec=myv)
        # handle hydrogen bonds
        if ssc in ['(','[','<']:
            lifo['x'].append(i)
        if ssc in [')',']','>']:
            j = lifo['x'].pop()
            graph.add_edge(i, j, label='=', type='basepair', len=1)

    # ADD BACKBONE
    lastgoodnode =  0
    for i in range(len(stru)-1):
        a,b = stru[i]=='.', stru[i+1]=='.'
        if a == b: # if a and b are the same we can just insert a normal edge
            graph.add_edge(i,i+1, label='-', type='backbone', len=1)
        elif a and not b: #  .-
            graph.add_edge(i,i+1, label='zz', nesting=True) #nesting are dark edges in eden
            if lastgoodnode:
                graph.add_edge(lastgoodnode, i+1, label='-', type='backbone', len=1)
        elif b and not a: #  -.
            lastgoodnode = i
            graph.add_edge(i,i+1, label='zz', nesting=True) #nesting are dark edges in eden
    return graph




def _most_common_nucs(ali):
    ''' goes through the alignments columnwise and finds the most common ACGU'''
    def getch(x):
        col = ali.alignment[:,x].tolist()
        counts = Counter( col )
        for k,v in counts.most_common():
            if k in "ACGU":
                return k
    return Map(getch, Range(ali.alignment.shape[1]))




def mainchainentropy(ali, structure  = 'SS_cons'):
    '''
        ATTENTION: . is sometimes in preserved columns so this is  bad-ish
        cleaning up the graph generation a bit,
        will keep a main alignment and just encode that together with the entropy...
        hopefully will keep a consensus to use for printing later..
        maybe i can also use sscons as vec-features
    '''
    graph = nx.Graph()
    lifo = defaultdict(list)
    nucs = _most_common_nucs(ali)
    simple_sscons = ali.gc[structure]
    simple_sscons = simple_sscons.replace(',',':')
    simple_sscons = simple_sscons.replace('-',':')
    simple_sscons = simple_sscons.replace('_',':')
    ali.gc[structure] = simple_sscons

    sequence = ''
    conSS = ''

    for i, (struct,nuc) in enumerate(zip(simple_sscons,nucs)):
        if struct != '.' and nuc != None: # ATTENTION! some structures have a :  but there is not even one nucleotide listed
            try:
                conSS += struct
                sequence+=nuc
                myv  = [ ord(ali.gc[k][i]) for k in ali.gc.keys() ]
                graph.add_node(i, label=nuc, vec=myv)
                # handle hydrogen bonds
                if struct in ['(','[','<']:
                    lifo['x'].append(i)
                if struct in [')',']','>']:
                    j = lifo['x'].pop()
                    graph.add_edge(i, j, label='=', type='basepair', len=1)
            except:
                print("ERROR IN FILE", ali.fname)

    # ADD BACKBONE
    nodes = list(graph)
    nodes.sort()
    for i in Range(len(nodes)-1):
            a,b = nodes[i], nodes[i+1]
            graph.add_edge(a,b, label='-', type='backbone', len=1)

    graph.graph = {}
    graph.graph['structure'] = conSS
    graph.graph['sequence'] = sequence
    return graph

import structout as so

import warnings
warnings.filterwarnings("ignore")

def scclust(ali):
    '''
    we try to recreate rna sc clust
    '''
    graph = nx.Graph()


    # adding nodes
    for i,(e,c) in enumerate(zip(ali.alignment[0], ali.covariance)):
        if e == f'.':
            continue
        symbol = e if c < .05 else f'C'
        graph.add_node(i, label=symbol)

    # adding structure
    for i,j in ali.pairs:
        if i in graph and j in graph:
            graph.add_edge(i, j, label='=', len=1)
        else:
            ali.skippedAPair = True
    # adding backbone
    nodes = list(graph.nodes())
    nodes.sort()
    for i in range(len(nodes)-1):
        graph.add_edge(nodes[i], nodes[i+1], label='-', len=1)

    # add fakenodes
    pairdict = dict(ali.pairs)
    for i in range(len(nodes)-1):
        # i and i+1 are definitely connected...
        i,j = nodes[i],nodes[i+1]
        a = pairdict.get(i,-1)
        b = pairdict.get(j,-1)
        dict_of_a_or_empty = graph._adj.get(a,{})
        if b in dict_of_a_or_empty:
            newnode = max(nodes)+1
            graph.add_node(newnode, label='o')
            graph.add_edges_from([(newnode,z) for z in [i,j,a,b]],label = '*')


    # so.gprint(graph, size = 30)
        #graph.add_edge(nodes[i], nodes[i+1], label='-', len=1)
    return graph



'''
def rfamseedfilebasic(ali, structure  = 'SS_cons'):

    graph = nx.Graph()
    lifo = defaultdict(list)
    if f'RF' not in ali.gc:
        print(f"{ ali.gf = }")
        print(f"{ ali.gc = }")
    nucs = ali.gc[f'RF'].upper()
    simple_sscons = ali.gc[structure]
    simple_sscons = simple_sscons.replace(',',':')
    simple_sscons = simple_sscons.replace('-',':')
    simple_sscons = simple_sscons.replace('_',':')
    ali.gc[structure] = simple_sscons

    sequence = ''
    conSS = ''

    for i, (struct,nuc) in enumerate(zip(simple_sscons,nucs)):
        if nuc != f'.': # ATTENTION! some structures have a :  but there is not even one nucleotide listed
            try:
                conSS += struct
                sequence+=nuc
                myv  = [ ord(ali.gc[k][i]) for k in ali.gc.keys() ]
                graph.add_node(i, label=nuc, vec=myv)
                # handle hydrogen bonds
                if struct in ['(','[','<']:
                    lifo['x'].append(i)
                if struct in [')',']','>']:
                    j = lifo['x'].pop()
                    graph.add_edge(i, j, label='=', type='basepair', len=1)
            except:
                print("ERROR IN FILE", ali.fname)

    # ADD BACKBONE
    nodes = list(graph)
    nodes.sort()
    for i in Range(len(nodes)-1):
            a,b = nodes[i], nodes[i+1]
            graph.add_edge(a,b, label='-', type='backbone', len=1)

    graph.graph = {}
    graph.graph['structure'] = conSS
    graph.graph['sequence'] = sequence
    return graph
'''


def clean_structure(struct):
    struct = struct.replace(',',':')
    struct = struct.replace('-',':')
    struct = struct.replace('_',':')
    struct = struct.replace('[','(')
    struct = struct.replace('<','(')
    struct = struct.replace('{','(')
    struct = struct.replace(']',')')
    struct = struct.replace('>',')')
    struct = struct.replace('}',')')
    return  struct

letterpairs = { chr(97+i):chr(65+i) for i in range(26) }

def  rfam_clean(ali):

    graph = nx.Graph()
    lifo = []
    lifo_pseudoknots = defaultdict(list)
    assert 'RF' in ali.gc

    ali.gc['clean_structure'] =  clean_structure (ali.gc['SS_cons'])


    sequence = ''
    conSS = ''
    ali.pseudoknot = any([a in letterpairs.keys() for a in ali.gc['clean_structure']])

    for i, (struct,nuc) in enumerate(zip(ali.gc['clean_structure'],ali.gc['RF'].upper())):
        if nuc != '.': # ATTENTION! some structures have a :  but there is not even one nucleotide listed
            try:
                conSS += struct
                sequence+=nuc
                # myv  = [ ord(ali.gc[k][i]) for k in ali.gc.keys() ]
                graph.add_node(i, label=nuc, vec=[])
                # handle hydrogen bonds
                if struct  == '(':
                    lifo.append(i)
                if struct == ')':
                    j = lifo.pop()
                    graph.add_edge(i, j, label='=', type='basepair', len=1)
                if ali.pseudoknot:
                    if struct in letterpairs.values():
                        lifo_pseudoknots[struct].append(i)
                    if struct in letterpairs.keys():
                        j = lifo_pseudoknots[letterpairs[struct]].pop()
                        graph.add_edge(i, j, label='=', type='basepair', len=1)

            except Exception as e:
                print("ERROR IN FILE", ali.fname,e)
                print (letterpairs, ali.gc['clean_structure'])

    # ADD BACKBONE
    nodes = list(graph)
    nodes.sort()
    for i in Range(len(nodes)-1):
            a,b = nodes[i], nodes[i+1]
            graph.add_edge(a,b, label='-', type='backbone', len=1)
    graph.graph = {}
    graph.graph['structure'] = conSS
    graph.graph['sequence'] = sequence

    ali.graph = graph
    ali.graphnodes = list(graph)

    return ali


def normalize(ctr):
    ctr.pop('-',None)
    su = sum(ctr.values())
    return [(k, v/su) for k,v in ctr.most_common()]


'''
def rfam_graph_structure_deco(ali):
    nuc_distribution = {i:normalize(Counter(ali.alignment[:,i])) for i in list(ali.graph)}

    def dist2vec(nudi):
        nudi = dict(nudi)
        nudi['R'] = sum([nudi.get(x,0) for x in 'A G'.split()])
        nudi['Y'] = sum([nudi.get(x,0) for x in 'C U'.split()])
        levels =  np.array([nudi.get(x,0) for x in f"A G C U Y R".split()])
        thresh = [.25,.5,.75,.95]
        print(f"{nudi=}")
        return np.array([ e > t for e in levels for t in thresh])

    for n in ali.graph:
        ali.graph.nodes[n]['label'] = '1'
        ali.graph.nodes[n]['vec'] = dist2vec(nuc_distribution[n])

    return ali

'''

def rfam_graph_decoration(ali, RY_thresh = .6, nuc_thresh = .85,
                          conservation = [.50,.75,.90,.95],
                          covariance = False,
                          nucleotideRNAFM = False ,
                          sloppy = False,
                          progress = 0,
                          cons_raw = False,
                          fake_nodes = False):
    vec = []
    for n in ali.graph.nodes:
        ali.graph.nodes[n]['vec'] = []
    nuc_distribution = {i:normalize(Counter(ali.alignment[:,i])) for i in list(ali.graph)}

    if RY_thresh or nuc_thresh:
        set_base_label(ali,nuc_distribution, nuc_thresh= nuc_thresh,RY_thresh= RY_thresh)

    if conservation:
        vec+=[0]*len(conservation)
        vec_add_conservation(ali, nuc_distribution, conservation)

    if covariance:
        set_base_covariance(ali, covariance)

    if sloppy:
        vec+=[0]*2
        vec_add_sloppy(ali)

    if progress:
        v_used = vec_add_progress(ali,progress)
        vec += [0]*v_used
    if cons_raw:
        vec += [0]
        vec_add_cons_raw(ali, nuc_distribution)

    if nucleotideRNAFM:
        vec += [0]*620
        vec_add_nucleotideembedding(ali)

    if fake_nodes:
        # add this last so that we know how long the vecs of the default nodes are
        add_fake_nodes(ali, vec)

    #if 'vec' in ali.graph.nodes[list(ali.graph.nodes)]:
    for n in ali.graph.nodes:
        ali.graph.nodes[n]['vec'] = np.array(ali.graph.nodes[n]['vec'])

    return ali


def vec_add_progress(ali, progress):
    proglen = progress if progress > 1 else 25
    nodes = list(ali.graph)
    for i,n in enumerate(nodes):
        prog = i/len(nodes)
        prog = int(prog*proglen)
        base = [0]*proglen
        base [prog] = 1
        ali.graph.nodes[n]['vec'] += base
    return proglen


def vec_add_cons_raw(ali, nuc_distribution):
    nodes = list(ali.graph)
    for n in nodes:
        add = 0
        if ali.graph.nodes[n]['label'] in 'ACGU':
            add = nuc_distribution[n][1]
        if ali.graph.nodes[n]['label'] == 'Y':
            nf = dict(nuc_distribution[n])
            add = nf['C'] + nf['U']
        if ali.graph.nodes[n]['label'] == 'R':
            nf = dict(nuc_distribution[n])
            add = nf['A'] + nf['G']
        ali.graph.nodes[n]['vec'] += [ add ]


def add_fake_nodes(ali,vec):
    # add fakenodes
    pairdict = dict(Flatten([[(a, b), (b, a)] for a, b, data in ali.graph.edges(data=True) if data['label'] == '=']))
    nodes = list(ali.graph)
    for n in range(len(nodes) - 1):
        # i and i+1 are definitely connected...
        i, j = nodes[n], nodes[n + 1]
        a = pairdict.get(i, -1)
        b = pairdict.get(j, -1)
        dict_of_a_or_empty = ali.graph._adj.get(a, {})
        if b in dict_of_a_or_empty and i < a:  # i<a prevents foring it twice
            newnode = max(ali.graph) + 1
            ali.graph.add_node(newnode, label='o', vec=[])
            ali.graph.add_edges_from([(newnode, z) for z in [i, j, a, b]], label='*')
            if len(vec) > 0:
                ali.graph.nodes[newnode]['vec'] = vec


def vec_add_sloppy(ali):
    for i in list(ali.graph):
        ali.graph.nodes[i]['vec'] += [0, 0]
    if '(' in ali.alignment:
        print(ali.fname)
    for a, b, data in ali.graph.edges(data=True):

        if data['label'] == '=':
            for z in zip(ali.alignment[:, a], ali.alignment[:, b]):

                if sorted(z) == list('GU'):
                    ali.graph.nodes[a]['vec'][-2] = 1
                    ali.graph.nodes[b]['vec'][-2] = 1
                    break

            for z in zip(ali.alignment[:, a], ali.alignment[:, b]):
                if sorted(z) == list('CU') or sorted(z) == list('AG') or sorted(z) == list('AC'):
                    ali.graph.nodes[a]['vec'][-1] = 1
                    ali.graph.nodes[b]['vec'][-1] = 1
                    break


def set_base_covariance(ali, covariance):
            for a,b,v in ali.rscape:
                    if v < covariance:
                        if a in ali.graph and b in ali.graph and  ali.graph.has_edge(a,b):
                            ali.graph.nodes[a]['label']= 'N'
                            ali.graph.nodes[b]['label']= 'N'
                        else:
                            pass #print(f"cov missmatch:{ ali.fname} {a} {b}")

def vec_add_conservation(ali, nuc_distribution, conservation):
    for i in list(ali.graph):
        nf = {a: 0 for a in f"ACGU"}
        nf.update(dict(nuc_distribution[i]))
        label = ali.graph.nodes[i]['label']
        percentage = nf[label] if label in 'ACGU' else (nf['A'] + nf['G'] if label == 'R' else nf['C'] + nf['U'])
        cons = [int(percentage > x) for x in conservation]
        ali.graph.nodes[i]['vec'] += cons


def set_base_label(ali, nuc_distribution,  nuc_thresh = .85 , RY_thresh = .6):
    thresh = nuc_thresh
    bonus = RY_thresh
    def chooselabel(nudi):
        nudid = dict(nudi)
        # return max(nudi.items(), key = lambda x:x[1])[0]
        k,v =  nudi[0]
        if v > thresh:
             return k

        if sum([nudid.get(x,0) for x in 'A G'.split()]) > bonus:
            return f'R'
        if sum([nudid.get(x,0) for x in 'C U'.split()]) > bonus:
            return f"Y"

        return f'0'

    for n in ali.graph:
        ali.graph.nodes[n]['label'] = chooselabel(nuc_distribution[n])
        # ali.graph.nodes[n]['vec'] = dist2vec(nuc_distribution[n])
    return ali



def vec_add_nucleotideembedding(ali):
    # assert len(ali) == 408, 'there are files called 0..407'
    vectors = np.load(f"/home/ubuntu/RNAEMBED/RNA-FM/redevelop/results/representations/{ali.gf['AC'].split()[1]}.npy")
    for n,vec in zip(list(ali.graph), vectors):
        ali.graph.nodes[n]['vec'] += vec.tolist()
    return ali


import copy, random
def manifest_subgraph(ali,n):
    ali = copy.deepcopy(ali)
    for nid in list(ali.graph):
        ali.graph.nodes[nid]['label'] = ali.alignment[n][nid]
    return ali

def manifest_subgraphs(a_l,maxgraphs = 100):
    ali, label = a_l
    ali_n_graphs = ali.alignment.shape[0]
    graphs_to_manifest = Range(ali_n_graphs)
    if ali_n_graphs > maxgraphs:
        random.shuffle(graphs_to_manifest)
        graphs_to_manifest= graphs_to_manifest[:maxgraphs]
    return [(manifest_subgraph(ali,x),label) for x in graphs_to_manifest ]

def manifest_sequences(alignments, labels, instances = 10, mp = False):
    mapper = ut.xmap if mp else Map
    a,labels = Transpose( Flatten ( mapper(manifest_subgraphs , zip(alignments,labels),
                                                maxgraphs = instances)))
    return a, labels





import networkx as nx
import numpy as np
import structout as so
from lmz import Range, Map
from matplotlib import pyplot as plt
from structout.rnagraph import RNAprint

import yoda.alignments

def scatter(X, y):
    plt.scatter(*X.T, c = y)
    plt.show()
    plt.close()


def asciigraph(ali):
    graph = nx.relabel_nodes(ali.graph,mapping = dict(zip(ali.graphnodes, Range(ali.graphnodes))),copy = True)
    RNAprint(graph, size = 2)
    return ali.graph


def shownodes(ali):
    for n in ali.graph.nodes:
        print(ali.graph.nodes[n])


def plotneighbors(alignments,distance_matrix, alignment_id = 5, num = 3):
    def plotalign(a1,a2):
        # _fancyalignment(seq1,seq2,str1,str2)
        _fancyalignment(a1.graph.graph['sequence'],
                        a2.graph.graph['sequence'],
                        a1.graph.graph['structure'],
                        a2.graph.graph['structure'])

    for i in np.argsort(distance_matrix[alignment_id])[1:num+1]:
        plotalign(alignments[alignment_id], alignments[i])


def plot_NN(dists, classes=0):
    NN = dists[:,1]
    if isinstance(classes, int):
            so.hist(NN,bins= 40)
    else:
        for e in np.unique(classes):
            so.hist(NN[classes == e],bins= 40)
    print(f"closest {min(NN)}")


def _lsearch(seq):
    # returns index of first char
    for i,e in enumerate(seq):
        if e != '-':
            return i


def _rsearch(seq):
    numfucks = _lsearch(seq[::-1])
    return len(seq) - numfucks


def _fancyalignment(seq1,seq2,str1,str2):

    al1, al2 = yoda.pairwise_alignments.needle(seq1, seq2)

    def adjuststruct(al1,str1):
        # 1. insert dashes into the str
        str1=list(str1)
        re = ''
        for e in al1[::-1]:
            if e == '-':
                re+=e
            else:
                re+=str1.pop()
        return re[::-1]

    str1 =  adjuststruct(al1,str1)
    str2 =  adjuststruct(al2,str2)


    '''
    here we have st1+2 and al1+2 that all have the same lengh
    ... next we want to format it...
    everything is aligned, we just cut it into 3 pieces
    '''

    # cutoff
    left = max(Map(_lsearch, (al1,al2)))
    right = min(Map(_rsearch, (al1,al2)))


    def cutprint(item):
        nust = item[left:right]
        left1  = _countfmt(item[:left])
        right1  = _countfmt(item[right:])
        nust = f'{left1} {nust} {right1}'
        print(nust)

    Map(cutprint,(al1,str1,al2,str2))
    print()
    print()


def _countfmt(item):
    # count '-' and gives the number a pading so the string has length 4
    cnt = str(len([a for a in item if a !='-']))
    return cnt+' '*(4-len(cnt))


def test_fancy_alignment():
    ab = ('----asdasda asd as da sd---', '---asdasdasd asd a sd---', )
    _fancyalignment(*ab,*ab)



from lmz import Map
from sklearn.metrics import silhouette_score
import numpy as np
from ubergauss import tools as ut
from sklearn.metrics import pairwise
from yoda import alignments
from yoda.ml.pairwise_alignments import smith_waterman_score, needle
from eden import sequence as eseq



import structout as so
def smithwaterman(ali, labels):
    # ali = ali[:10]
    # labels = labels[:10]
    l= len(ali)
    m = np.ndarray((l,l))
    tasks = [(i,j) for i in range(l) for j in range(i+1,l)]
    def sim(task):
        a,b = ali[task[0]], ali[task[1]]
        sa = a.graph.graph[f'sequence']
        sb = b.graph.graph[f'sequence']
        # return 1-(smith_waterman_score(sa,sb)/min(len(sa), len(sb)))
        return smith_waterman_score(sa,sb)

    res = ut.xmap(sim, tasks)
    for r,(i,j) in zip(res, tasks):
        m[i,j] = r
        m[j,i] = r


    return m

def score_smithwaterman(ali, labels):
    m = smithwaterman(ali, labels)
    so.heatmap(m)
    m *=-1
    m -= np.min(m)
    np.fill_diagonal(m,0)
    so.heatmap(m)
    return silhouette_score(m, labels, metric='precomputed')



def needleman(ali, labels):
    # ali = ali[:10]
    # labels = labels[:10]
    l= len(ali)
    m = np.ndarray((l,l))
    tasks = [(i,j) for i in range(l) for j in range(i,l)]
    def sim(task):
        a,b = ali[task[0]], ali[task[1]]
        sa = a.graph.graph[f'sequence']
        sb = b.graph.graph[f'sequence']
        #normalizing is pointless...
        return 1-(needle(sa,sb, get_score_only=True)/min(len(sa), len(sb)))

    res = ut.xmap(sim, tasks)
    for r,(i,j) in zip(res, tasks):
        m[i,j] = r
        m[j,i] = r

    return silhouette_score(m, labels, metric='precomputed')



def gapped_kmers(ali, labels):
    ali = eseq.vectorize([a.graph.graph['sequence'] for a in ali])
    m = pairwise.cosine_distances(ali)
    return silhouette_score(m, labels, metric='precomputed')


from yoda.graphs import vectorize_alignments
from yoda.ml import get_distance_method
def simple_graph( alignments,labels):
    vectors = vectorize_alignments(alignments, mp=True)
    # di = get_distance_method(f'linkern')(vectors)
    return silhouette_score(vectors, labels)


if __name__ == f"__main__":
    ali, label = alignments.load_rfam()
    #ali = ali[:5]
    #label = label[:5]
    data = ali,label
    print(f"{needleman(*data)=}")
    print(f"{score_smithwaterman(*data)=}")

    # print(f"{gapped_kmers(*data)=}")
    # print(f"{simple_graph(*data)=}")




import numpy as np


# def getlabels(ali):
#     '''
#     matches the ID field in the alignments with the "clans"-cluster-list
#     returns cluster-labels
#     '''
#     y = np.zeros(len(ali),dtype=int)
#     for i,e in enumerate(clans.split(f'\n')):
#         rnaid = e.split()[1:]
#         for j,a in enumerate( ali):
#             if a.gf[f'ID'][3:] in rnaid:
#                y[j] = i + 1
#     return y



def get_label_dictionary():
    d = {}
    for i,e in enumerate(clans.split(f'\n')):
        rna_ids = e.split()[1:]
        for rna_id in rna_ids:
            d[rna_id] = i+1
    return d

def getlabels(ali):
    '''
    matches the ID field in the alignments with the "clans"-cluster-list
    returns cluster-labels
    '''

    label_dict = get_label_dictionary()
    print(f"{ label_dict=}")
    y = np.zeros(len(ali),dtype=int)
    for j,a in enumerate( ali):
        label = label_dict.get( a.gf[f'ID'][3:] ,  0 )
        y[j] = label
    return y


# these clans are too long for the NEURAL NETWORK Node embedder RNA-FM
#CL00112	5_8S_rRNA	LSU_rRNA_archaea	LSU_rRNA_bacteria	LSU_rRNA_eukarya	LSU_trypano_mito
#CL00111	SSU_rRNA_bacteria	SSU_rRNA_archaea	SSU_rRNA_eukarya	SSU_rRNA_microsporidia	SSU_trypano_mito
#CL00004	Telomerase-vert	Telomerase-cil	Sacc_telomerase	Telomerase_Asco


clans = '''CL00110	mir-19	mir-363
CL00071	SNORD88	snR76	snoR118
CL00070	snosnR60_Z15	SNORD77	Afu_263
CL00073	snoR30	SNORD100
CL00072	SNORD96	SNORD2
CL00075	SNORD105	snoU105B
CL00074	SNORD101	snoR60
CL00077	SNORND104	snR58
CL00076	SNORD110	snoR14
CL00079	snR68	snoR27
CL00078	SNORA73	snR30
CL00068	SNORD62	snoR41	snR41
CL00069	SNORD79	SNORD113	SNORD74	snoR44_J54	snosnR64	snoMe28S-Cm2645	SNORD64	SNORD112	SNORD109A
CL00066	SNORD60	snoR1	snosnR48	snoMe28S-G3255	sn2903
CL00067	SNORD61	snoU61	SNORD125
CL00064	SNORD58	SNORD99
CL00065	snoZ159	SNORD59	snosnR54
CL00062	SNORD49	snoZ112	snoU49
CL00063	SNORD52	SNORD53_SNORD92	snoZ157	snR78
CL00060	SNORD44	snoZ102_R77
CL00061	SNORD46	snR63
CL00093	mir-183	mir-182	mir-263	mir-228
CL00092	mir-137	mir-234
CL00091	bantam	mir-81
CL00090	mir-74	mir-73
CL00097	mir-BART1	mir-BART3
CL00096	mir-515	mir-290	mir-302	miR-430
CL00095	mir-279	mir-996
CL00094	mir-216	mir-283
CL00099	MIR171_1	MIR171_2
CL00098	MIR169_2	MIR169_5
CL00019	SCARNA4	SCARNA15
CL00018	SCARNA24	SCARNA3
CL00013	7SK	Arthropod_7SK
CL00012	SAM	SAM-IV	SAM-I-IV-variant
CL00011	GlmZ_SraJ	GlmY_tke1
CL00010	Hammerhead_3	Hammerhead_1	Hammerhead_HH9	Hammerhead_II	Hammerhead_HH10
CL00017	IRES_HCV	IRES_Pesti
CL00016	FinP	traJ_5
CL00015	CRISPR-DR5	CRISPR-DR7	CRISPR-DR63	CRISPR-DR64
CL00014	CRISPR-DR2	CRISPR-DR4	CRISPR-DR14	CRISPR-DR17	CRISPR-DR25	CRISPR-DR43	CRISPR-DR66
CL00080	snoR53	snoR53Y
CL00081	snoZ279_R105_R108	snoU13
CL00082	snoU89	snoU85
CL00083	mir-6	mir-11
CL00084	mir-3	mir-318
CL00085	mir-16	mir-15
CL00086	mir-28	mir-708
CL00087	mir-34	mir-449
CL00088	mir-36	mir-42	mir-35
CL00089	mir-190	mir-50
CL00118	L31-Coriobacteria	L31-Corynebacteriaceae	L31-Firmicutes	L31-Gammaproteobacteria	L31-Actinobacteria
CL00008	U54	snoU54
CL00009	U6	U6atac
CL00125	Glycine	glycine-GGAnGA
CL00124	TD-1	TD-2
CL00123	Purine	MFR	2dG-II
CL00122	NiCo	NiCo-AnGGAG
CL00121	S15	S15-Methanomicrobia	S15-Halobacteria	S15-Flavobacteria
CL00120	Twister-P5	twister-P3	twister-P1
CL00001	tRNA	tmRNA	alpha_tmRNA	beta_tmRNA	cyano_tmRNA	tRNA-Sec	mt-tmRNA
CL00002	RNaseP_nuc	RNaseP_bact_a	RNaseP_bact_b	RNase_MRP	RNaseP_arch	RNase_P	RNaseP-T
CL00003	Metazoa_SRP	Bacteria_small_SRP	Fungi_SRP	Dictyostelium_SRP	Bacteria_large_SRP	Plant_SRP	Protozoa_SRP	Archaea_SRP	Bacteroidales_small_SRP
CL00005	U1	U1_yeast	U11	Gl_U1
CL00006	U2	U12
CL00007	U4	U4atac
CL00035	SNORA36	snR36	snR44	S_pombe_snR36
CL00034	SNORA50	SNORA54	SNORA35	SNORA76
CL00037	SNORA48	snR86
CL00036	SNORA44	SNORA58	snR161	snR9
CL00031	SNORA21	snR10	S_pombe_snR10
CL00030	SNORA20	SNORA29
CL00033	SNORA28	snopsi18S-841	snR80
CL00032	SNORA27	SNORA26	snR42	S_pombe_snR42
CL00039	SNORA56	snR8
CL00038	SNORA66	snoR98	SNORA52	SNORA18	snoR80	snR49	snR5	S_pombe_snR5
CL00131	Flavi_ISFV_repeat_Ra_Rb	Flavi_ISFV_repeat_Rb	Flavi_ISFV_repeat_Ra
CL00132	snoR9	snoPyro_CD	sR47	sR60
CL00127	c-di-GMP-II	c-di-GMP-II-GAG
CL00126	c-di-GMP-I	c-di-GMP-I-GGC	c-di-GMP-I-UAU
CL00022	SNORA3	snR11
CL00023	SNORA4	snR83
CL00020	SL1	SL2
CL00021	SNORA16	SNORA2	snopsi28S-3327	snR189	snR46	S_pombe_snR46
CL00026	SNORA8	snR31
CL00027	SNORA9	snR33	snR43	S_pombe_snR33
CL00024	SNORA5	snR85
CL00025	SNORA7	snR81
CL00028	SNORA13	snR35	S_pombe_snR35
CL00029	SNORA43	SNORA17
CL00101	Cobalamin	AdoCbl_riboswitch	AdoCbl-variant
CL00100	U3	Fungi_U3	Plant_U3	ACEA_U3
CL00103	SNORD11	SNORD11B
CL00102	group-II-D1D4-1	group-II-D1D4-2	group-II-D1D4-3	group-II-D1D4-4	group-II-D1D4-5	group-II-D1D4-6	group-II-D1D4-7
CL00105	SraC_RyeA	SdsR_RyeB
CL00104	HBV_epsilon	AHBV_epsilon
CL00106	CsrB	CsrC	McaS	PrrB_RsmZ	RsmY	TwoAYGGAY	rsmX	Ysr186_sR026_CsrC	RsmW
CL00108	suhB	ar15
CL00129	cHP	DENV_SLA	Flavivirus-5UTR
CL00128	C4	c4-2	c4-a1b1
CL00057	SNORD39	snoZ7	snoZ101	SNORD65	snoR77Y	snR77
CL00056	SNORD35	snoZ161_228	snR73
CL00055	SNORD34	snR62
CL00054	SNORD33	snoZ196	SNORD51	snosnR55	snoMe18S-Um1356	snoMe28S-Am982	snR39	snR40
CL00053	SNORD31	snoZ17	snR67	snoR35
CL00052	SNORD30	snoU30
CL00051	SNORD36	SNORD29	snoZ223	SNORD38	snosnR69	snosnR61	snosnR71	SNORD78	snoR69Y	snR47	snoU36a
CL00050	SNORD26	SNORD81
CL00116	aCoV-5UTR	bCoV-5UTR	gCoV-5UTR	dCoV-5UTR	Sarbecovirus-5UTR
CL00117	s2m	Corona_pk3	aCoV-3UTR	bCoV-3UTR	gCoV-3UTR	dCoV-3UTR	Sarbecovirus-3UTR
CL00114	LhrC	rli22	rli33
CL00115	DUF805b	DUF805
CL00113	5S_rRNA	mtPerm-5S
CL00059	SNORD43	snR70
CL00058	SNORD57	SNORD41	snR51
CL00044	SNORD12	snR190
CL00045	SNORD15	snR75	snR13	snoZ5
CL00046	SNORD16	snR87
CL00047	SNORD18	snoU18
CL00040	SNORA62	snR3	snR82	S_pombe_snR3
CL00041	SNORA64	snR37
CL00042	SNORA65	snR34	snoR2
CL00043	SNORA74	snR191
CL00119	S4-Fusobacteriales	S4-Bacteroidia	S4-Clostridia	S4-Flavobacteria
CL00048	SNORD19	SNORD19B
CL00049	SNORD25	snR56f'''



from lmz import Zip, Range,Transpose
import numpy as np
from ubergauss import tools as ut
from yoda.ml.simpleMl import overlap_coef
from yoda.alignments import ali2graph
from yoda.draw import _fancyalignment

'''
use the vectors to make a nearest neighbor model...
there is a filter to remove stuff that is too similar
'''


def neighbors(vecz, k = 100):
    def doarow(x):
        distances = np.array([overlap_coef(vecz[x], vecz[y]) for y in Range(vecz)])

        if k < 1:
            return distances, 0

        indices_k = np.argsort(distances)[:k]
        # fix order such that first id is the instance itself
        # note distances dont matter because if we are not first the distance is zero
        if indices_k[0] != x:
            other = np.where(indices_k == x)[0][0]
            indices_k[0], indices_k[other]= indices_k[other], indices_k[0]
        dist_k = [distances[ar] for ar in indices_k]

        return dist_k,indices_k

    d,i  = Transpose ( ut.xmap(doarow, Range(vecz)))
    d = np.array(d)
    i = np.vstack(i)
    return d,i





def _plot_alignments(filelist):
    #!cat {files[b]}
    #!cat {files[a]}
    file2graph = lambda x: ali2graph.mainchainentripy(vv.read_stk_file(x)[0])
    graphs = [file2graph for f in filelist]
    # ed.draw_graph_set(graphs, vertex_border=False, vertex_size=200)
    #plt.show()
    #so.gprint(graphs, size = 40)

    # for f in graphs:
    #     print(f.graph.get('sequence', ''))
    #     print(f.graph.get('structure', ''))

    seq1 = graphs[0].graph.get('sequence', '')
    str1 = graphs[0].graph.get('structure', '')
    seq2 = graphs[1].graph.get('sequence', '')
    str2 = graphs[1].graph.get('structure', '')

    _fancyalignment(seq1,seq2,str1,str2)


def _sortbythresh(distances, indices):
    dst = [(d,s,t) for s,(d,t) in enumerate(zip(distances[:,1],indices[:,1]))] #distance source target
    dst.sort()
    return dst


def plotbythresh(distances, indices, files, thresh = .7, max = 5):
    i = 0
    dst = _sortbythresh(distances, indices)
    for d,a,b in dst:
        if d > thresh:
            if i % 2 == 0: # they al ways show up in pairs, drawing onece is enough
                print(f"{d} {files[a]} {files[b]}")
                _plot_alignments([files[z] for z in [a,b]])
            i+=1
            if i > max*2:
                break


##############
# doing some filtering
###############


def filter_thresh(distances, indices, thresh = .7, checkmax = 400):
    rmlist = set()
    for d, i in zip(distances, indices):
        for zz, (other,dist) in enumerate(Zip(i[1:],d[1:])[:checkmax]):
            if dist > thresh:
                break
            if dist < thresh and (other not in rmlist):
                rmlist.add(i[0])
                break
    return rmlist



def filter_dump(rmlist, files, out = 'okfiles.json'):
    filesF = [f for i,f in enumerate(files) if i not in rmlist]
    ut.jdumpfile(filesF,out)





from lmz import Map,Zip,Filter,Grouper,Range,Transpose,Flatten
from yoda import alignments
from sklearn.metrics import silhouette_score
import numpy as np

from yoda.alignments import alignment

def score(mtx,data):
    labels = data[1]
    return silhouette_score(mtx, labels, metric='precomputed')

def data_to_reffile(data):
    alis,_ = data
    reflist = [ ali.gf[f'AC'].split()[1] for ali in alis]
    with open(f'reffile.delme',f'w') as f:
        f.write(f'\n'.join(reflist))
    return reflist


def data_to_fasta(data):
    alis,_ = data
    lines = []

    # sequences = [ ali.graph.graph['sequence'] for ali in alis]
    # for i,s in enumerate(sequences):
    #     lines.append( f'>{i}')
    #     lines.append( s)

    for ali in alis:
        rfamid = ali.gf['AC'].split()[1]
        lines.append( f'>{rfamid}')
        lines.append( ali.graph.graph['sequence'])

    with open(f'fasta.delme',f'w') as f:
        f.write(f'\n'.join(lines))
    return lines

from ubergauss import tools as ut
def readCmscanAndMakeTable(data):
    reflist = data_to_reffile(data)
    refdict = {nr:idx for idx,nr in enumerate(reflist)}
    l = len(refdict)
    distmtx= np.ones((l,l))
    distmtx*=0
    for  line in open(f'delme.tbl',f'r').readlines():
        if not line.startswith(f"#"):
            line = line.split()
            if line[1] not in refdict:
                continue
            x = refdict[line[1]]
            #y = int(line[2])
            y = refdict[line[2]]
            evalue = float(line[15])
            distmtx[x,y] = evalue
            distmtx[y,x] = evalue
            # print(f"{ evalue=}")
    np.fill_diagonal(distmtx,0)
    return distmtx
from yoda import ml, draw


import matplotlib
matplotlib.use('module://matplotlib-backend-sixel')
from matplotlib import pyplot as plt



from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics.cluster import adjusted_rand_score, rand_score
def eval_agglo_ari(dist,labels):
    rand_indices = []
    adjusted_rand_indices = []
    for n in np.unique(dist):
        predict = AgglomerativeClustering(n_clusters = None,linkage='single',
                    distance_threshold=n,affinity = 'precomputed').fit_predict(dist)
        adjusted_rand_indices.append(adjusted_rand_score(predict, labels))
        rand_indices.append(rand_score(predict, labels))

    print(f"{max(rand_indices)=}")
    print(f"{max(adjusted_rand_indices)=}")




if __name__ == f"__main__":
    data = alignments.load_rfam()


    # data_to_fasta(data)
    # data_to_reffile(data)

    '''
    so now we need to get the dist matrix...
    cmfetch --index Rfam.cm
    cmfetch -f Rfam.cm reffile.delme > reffile.delme.cm
    cmpress reffile.delme.cm
    cmcalibrate reffile.delme.cm
    cmscan -g -E 10000 --noali --acc --tblout=delme.tbl  reffile.delme.cm fasta.delme
    '''
    dist = readCmscanAndMakeTable(data)
    eval_agglo_ari(dist,data[1])
    exit()

    import structout as so
    from scipy import sparse
    import seaborn as sns

    sns.clustermap(dist)
    plt.show()

    so.heatmap(sparse.csr_matrix(dist))

    em = ml.embed(dist)
    draw.scatter(em, data[1])


    # draw.scatter(em, Range(data[1]))
    # for e in np.logspace(1,-50):
    #     d2 = dist.copy()
    #     d2[d2==10] = e
    #     print(f"{e=}")
    #     print(f"{score(d2, data)=}")





import numpy as np

# Create a sample NumPy array with empty columns
arr = np.array([[1, 2, 0, 4],
                [5, 6, 0, 8],
                [0, 0, 0, 0],
                [0, 0, 0, 0]])

# Find empty columns by checking if all elements in each column are zeros
empty_columns = np.all(arr == 0, axis=0)

# Use boolean indexing to remove empty columns
result = arr[:, ~empty_columns]

# Print the result
print(result)



from ubergauss import tools as ut
from yoda.graphs import ali2graph
from yoda.alignments import filein, clans, alignment


def load_rfam(seedpath = '~/Rfam.seed.utf8', full = False, add_cov = '~/rfam/test2'):
    alignments = filein.readseedfile(ut.fixpath( seedpath))
    labels = clans.getlabels(alignments)

    if not full:
        oklabel = labels != 0
        alignments = [a for a,ok in zip(alignments,oklabel) if ok > 0]
        labels = labels[oklabel]
    if add_cov:
        alignments = filein.addcov_rfam(alignments, add_cov)
    alignments = ut.xmap(ali2graph.rfam_clean, alignments)

    #check_labels(alignments,labels)

    return alignments, labels


def check_labels(alignments, labels):
    for a,l in zip(alignments[:5], labels[:5]):
        print(f"{l=}  {a.gf[f'ID'][3:]=} {a.gf[f'AC'].split()[1:]=}")


def subsample(a,l, num = 10):
    return a[:num], l[:num]


if __name__ == "__main__":
    z = load_rfam()[0]
    z = [zz for zz in z if zz.pseudoknot]
    import structout as so
    for zz in z:
        try:
            #so.rnagraph.RNAprint(zz.graph)
            so.gprint(zz.graph, size = 70)
        except:
            pass

